import os.path
from fnmatch import fnmatch


def _recursive_path_match(path_list, pattern_list, match_subpath):
    while pattern_list and pattern_list[0] == "**":
        pattern_list.pop(0)

    if not pattern_list:
        return True

    if not path_list:
        return False

    for i in range(len(path_list)):
        if _path_match(path_list[i:], pattern_list, match_subpath):
            return True
    return False


def _path_match(path_list, pattern_list, match_subpath):
    if not path_list or not pattern_list:
        return False

    for i, pattern in enumerate(pattern_list):
        if pattern == "**":
            # recursive matcher.
            return _recursive_path_match(path_list[i:], pattern_list[i + 1:], match_subpath)
        elif len(path_list) <= i:
            return False
        elif not fnmatch(path_list[i], pattern):
                return False
    return True if match_subpath else len(path_list) == i + 1


def path_match(path, pattern, match_subpath=False):
    path = os.path.normpath(path)
    pattern = os.path.normpath(pattern)
    path_list = path.split(os.path.sep)
    pattern_list = pattern.split(os.path.sep)
    return _path_match(path_list, pattern_list, match_subpath)


def find_matching_pattern(path, patterns, match_subpath=False):
    for pattern in patterns:
        if path_match(path, pattern, match_subpath):
            return pattern
    return None


if __name__ == '__main__':
    assert path_match('a/b/c/d', 'a/b/c/d')
    assert not path_match('a/b/c/d', 'a/b/c')
    assert path_match('a/b/c/d', 'a/*/c/d')
    assert not path_match('a/b/c/d', 'a/*/d/d')
    assert path_match('a/b/c/d', 'a/**/d')
    assert path_match('a/b/c/d', '**/d')
    assert path_match('a/b/c/d', '**')
    assert not path_match('a/b/c/d', '**/e')
    assert path_match('a/b/c/d', 'a/**/b/c/d')
    assert path_match('a/b/e.y', '**/*.y')
    assert path_match("a/b/c", "a", match_subpath=True)



__version__ = "0.0.11"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = '(detached from v0.0.11)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = 'b8d1705f26ed0388d3f9e0bb70b099943b9a0b24'
__git_head_subject__ = 'buildout.cfg: changed isolated python version to v2.7.6.6'
__git_head_message__ = ''
__git_dirty_diff__ = ''



SETUP_INFO = dict(
    name = 'infi.pysync',
    version = '0.0.11',
    author = 'Arnon Yaari',
    author_email = 'arnony@infinidat.com',

    url = 'https://infinigit.infinidat.com/host/pysync',
    license = 'PSF',
    description = """rsync-like utility to help sync dev files to a machine""",
    long_description = """rsync-like utility to help sync dev files to a machine""",

    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Python Software Foundation License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],

    install_requires = ['docopt',
'infi.watchdog',
'paramiko',
'setuptools'],
    namespace_packages = ['infi'],

    package_dir = {'': 'src'},
    package_data = {'': []},
    include_package_data = True,
    zip_safe = False,

    entry_points = dict(
        console_scripts = ['pysync = infi.pysync:main'],
        gui_scripts = [],
        ),
)

if SETUP_INFO['url'] is None:
    _ = SETUP_INFO.pop('url')

def setup():
    from setuptools import setup as _setup
    from setuptools import find_packages
    SETUP_INFO['packages'] = find_packages('src')
    _setup(**SETUP_INFO)

if __name__ == '__main__':
    setup()




#!/usr/bin/env python
"""
Sync Python source files, removes orphan .pyc files and ignores .git directory (or other patterns)

Usage:
  pysync [options] [-s PATTERN...] [-t PATTERN...] TARGET
  pysync [options] [-s PATTERN...] [-t PATTERN...] SOURCE TARGET

Options:
  -h, --help         Show this help
  -v, --verbose      Be verbose
  -n, --dry-run      Don't do anything, report it instead
  -w, --watch        Keep watching for changed files/directories in SOURCE and sync them to TARGET
  -p, --python       Python-aware synchronization:
                       - Don't copy .pyc from SOURCE
                       - Remove .pyc files in TARGET when syncing .py files
                       - Remove .pyc files in TARGET when there's no .py file
  -s PATTERN --skip-source=PATTERN   Don't sync PATTERN from source.
                                     Accepts path patterns (e.g. *, ?, []) and ** as a recursive match-all.
                                     Works well with path hierarchy (e.g. a/b/c.ext), comparison starts from SOURCE.
  -t PATTERN --skip-target=PATTERN   Don't remove objects matching PATTERN from TARGET when they don't exist in SOURCE.
                                     Same rules as -s.
  -b INTERVAL --batch-interval=INTERVAL  Interval in seconds to sync [default: 1]
  -i IDENTITY --identity=IDENTITY    SSH identity file to use when authenticating remote host

Options you probably don't want to use:
  --no-preserve-time                 Don't sync mtime/atime (default: False)
  --size-only                        Sync by size only (default: False)
  --mtime-only                       Sync by modification time only (default: False)

  TARGET             A scp-like path: [user@]host[:port][:path]
"""
import os
import sys
import errno
import time
import stat
from fnmatch import fnmatch
from docopt import docopt
from watchdog.observers import Observer
from watchdog.events import (FileSystemEventHandler,
                             EVENT_TYPE_MODIFIED, EVENT_TYPE_MOVED, EVENT_TYPE_CREATED, EVENT_TYPE_DELETED)
from watchdog.utils import has_attribute
from Queue import Queue, Empty
from threading import Thread

from .pathmatch import find_matching_pattern
from .sftp import SyncSFTPClient

verbose = None
dry_run = False
source_ignore_patterns = [".git", "conf"]
target_ignore_patterns = ["bin", "eggs", "develop-eggs", "parts", "**/__version__.py*", ".cache"]
python_mode = False
compare_by_size = True
compare_by_mtime = True


def vprint(msg, *args, **kwargs):
    global verbose
    if verbose:
        print(msg.format(*args, **kwargs))


class SyncWorker(Thread):
    def __init__(self, sftp, queue, source_base_path, target_base_path, batch_interval=1.0):
        self.sftp = sftp
        self.queue = queue
        self.source_base_path = source_base_path
        self.target_base_path = target_base_path
        self.batch_interval = batch_interval
        super(SyncWorker, self).__init__()

    def stop(self):
        self.queue.put(None)

    def run(self):
        try:
            batch_start_time = time.time()
            batch = []
            shutdown = False
            while not shutdown:
                event = None
                try:
                    wait_time = max(0, batch_start_time + self.batch_interval - time.time())
                    event = self.queue.get(timeout=wait_time)
                    self.queue.task_done()
                    if not event:
                        shutdown = True
                except Empty:
                    pass

                now = time.time()
                if event:
                    batch.append(event)
                if (now - batch_start_time) >= self.batch_interval or shutdown:
                    if batch:
                        self._handle_batch(batch)
                    batch_start_time = now
                    batch = []
        except:
            import traceback
            traceback.print_exc()

    def _handle_batch(self, batch):
        paths_synced = set()
        for event in batch:
            if event.event_type == EVENT_TYPE_DELETED:
                self.sftp.rm_rf(self._to_target_path(event.src_path))
                if event.src_path in paths_synced:
                    paths_synced.remove(event.src_path)
                self._remove_pyc_file_if_needed(event.src_path)
            elif event.event_type in (EVENT_TYPE_MODIFIED, EVENT_TYPE_CREATED) and not event.src_path in paths_synced:
                if os.path.exists(event.src_path):
                    self.sftp.put(event.src_path, self._to_target_path(event.src_path))
                    paths_synced.add(event.src_path)
                    self._remove_pyc_file_if_needed(event.src_path)
                else:
                    vprint("not syncing {}, already deleted in source.", event.src_path)
            elif event.event_type == EVENT_TYPE_MOVED:
                if event.src_path in paths_synced:
                    paths_synced.remove(event.src_path)
                if event.dest_path in paths_synced:
                    paths_synced.remove(event.dest_path)
                self.sftp.rename(self._to_target_path(event.src_path), self._to_target_path(event.dest_path))
                self._remove_pyc_file_if_needed(event.src_path)
                self._remove_pyc_file_if_needed(event.dest_path)

    def _remove_pyc_file_if_needed(self, path):
        fname = os.path.basename(path)
        if python_mode and fnmatch(fname, "*.py"):
            vprint("removing .pyc since {} was deleted", path)
            self.sftp.rm_rf(self._to_target_path(path + "c"))

    def _to_target_path(self, path):
        return _convert_path(self.source_base_path, self.target_base_path, path)


class SyncHandler(FileSystemEventHandler):
    def __init__(self, queue, source_base_path, source_ignore_patterns, *args, **kwargs):
        self.queue = queue
        self.source_base_path = os.path.normpath(os.path.abspath(source_base_path))
        self.source_ignore_patterns = source_ignore_patterns
        super(SyncHandler, self).__init__(*args, **kwargs)

    def dispatch(self, event):
        try:
            paths = [event.src_path]
            if has_attribute(event, 'dest_path'):
                paths.append(event.dest_path)

            paths = [os.path.relpath(os.path.normpath(os.path.abspath(path)), start=self.source_base_path) for path in paths]
            event_type_to_name = {EVENT_TYPE_MOVED: "move", EVENT_TYPE_CREATED: "create",
                                  EVENT_TYPE_MODIFIED: "modify", EVENT_TYPE_DELETED: "delete"}
            vprint("local change type: {} paths: {}", event_type_to_name[event.event_type], paths)
            for path in paths:
                pattern = find_matching_pattern(path, self.source_ignore_patterns, match_subpath=True)
                if pattern:
                    vprint("ignoring change for path {}, pattern: {}", path, pattern)
                    return

            self.queue.put(event)
        except:
            import traceback
            traceback.print_exc()


def _sync_file(sftp, source_path, source_stat, remote_path):
    local_size = source_stat.st_size
    local_time = int(source_stat.st_mtime)
    remote_stat = sftp.stat_or_none(remote_path)
    if remote_stat and not stat.S_ISREG(remote_stat.st_mode):
        raise Exception("remote path {} should be a regular file, but it's not.".format(remote_path))
    remote_size = remote_stat.st_size if remote_stat else None
    remote_time = int(remote_stat.st_mtime) if remote_stat else None

    size_diff = local_size != remote_size if compare_by_size else False
    time_diff = local_time != remote_time if compare_by_mtime else False

    vprint("sync_file comparison for {} -> {}: size={} ({} <=> {}) time={} ({} <=> {})".format(source_path, remote_path,
           size_diff if compare_by_size else "<unused>", local_size, remote_size,
           time_diff if compare_by_mtime else "<unused>", local_time, remote_time))
    if size_diff or time_diff:
        sftp.put(source_path, remote_path)

        if python_mode and source_path.endswith(".py"):
            remote_pyc = remote_path + "c"
            if sftp.stat_or_none(remote_pyc):
                vprint("removing remote .pyc file {}", remote_pyc)
                sftp.rm_rf(remote_pyc)


def filtered_walk(path, ignore_patterns=[], listdir_func=os.listdir, stat_func=os.lstat,
                  listdir_attr_func=None, compare_path=""):
    path = os.path.normpath(path)
    base_path = os.path.relpath(path, compare_path) if compare_path else path

    entries = []

    def _filter_entries(all_entries):
        for entry in all_entries:
            p = os.path.join(base_path, entry)
            pattern = find_matching_pattern(p, ignore_patterns, match_subpath=True)
            if not pattern:
                entries.append(entry)
            else:
                vprint("ignoring {}, ignore pattern: {}", p, pattern)

    if listdir_attr_func:
        stats = listdir_attr_func(path)
        _filter_entries(s.filename for s in stats)
        entry_set = set(entries)
        stats = [s for s in stats if s.filename in entry_set]
    else:
        _filter_entries(listdir_func(path))
        stats = [stat_func(os.path.join(path, e)) for e in entries]

    entry_to_stat = dict(zip(entries, stats))

    yield (path, entry_to_stat)

    for d in [e for e, s in entry_to_stat.items() if stat.S_ISDIR(s.st_mode)]:
        try:
            stat_func(os.path.join(path, d))  # just to make sure it still exists
            for result in filtered_walk(os.path.join(path, d), ignore_patterns, listdir_func, stat_func,
                                        listdir_attr_func, compare_path):
                yield result
        except (OSError, IOError), e:
            if e.errno != errno.ENOENT:
                raise


def _convert_path(from_base_path, to_base_path, path):
    relpath = os.path.relpath(path, start=from_base_path)
    return os.path.normpath(os.path.join(to_base_path, relpath))


def _add_source_to_target(sftp, source_base_path, remote_base_path):
    # Try to see which files/directories are missing remotely or have different size and create/copy.
    for (dirpath, entry_to_stat) in filtered_walk(source_base_path, ignore_patterns=source_ignore_patterns,
                                                  compare_path=source_base_path):
        remote_dirpath = _convert_path(source_base_path, remote_base_path, dirpath)
        sftp.mkdir_if_not_exist(remote_dirpath)

        for dirname in [k for k, s in entry_to_stat.items() if stat.S_ISDIR(s.st_mode)]:
            sftp.mkdir_if_not_exist(os.path.join(remote_dirpath, dirname))

        for fname, f_stat in [(f, s) for f, s in entry_to_stat.items() if stat.S_ISREG(s.st_mode)]:
            if not python_mode or not fnmatch(fname, '*.pyc'):
                _sync_file(sftp, os.path.join(dirpath, fname), f_stat, os.path.join(remote_dirpath, fname))


def _remove_target_when_missing_source(sftp, source_base_path, remote_base_path):
    for (dirpath, entry_to_stat) in filtered_walk(remote_base_path, ignore_patterns=target_ignore_patterns,
                                                  listdir_attr_func=sftp.listdir_attr, stat_func=sftp.stat,
                                                  compare_path=remote_base_path):
        src_dirpath = _convert_path(remote_base_path, source_base_path, dirpath)
        if not os.path.exists(src_dirpath):
            sftp.rm_rf(dirpath)
        else:
            for dirname in [k for k, s in entry_to_stat.items() if stat.S_ISDIR(s.st_mode)]:
                if not os.path.exists(os.path.join(src_dirpath, dirname)):
                    sftp.rm_rf(os.path.join(dirpath, dirname))

            for fname, f_stat in [(f, s) for f, s in entry_to_stat.items() if stat.S_ISREG(s.st_mode)]:
                if python_mode and fnmatch(fname, '*.pyc'):
                    # Make sure there's a .py file in the source.
                    if not os.path.exists(os.path.join(src_dirpath, fname[:-1])):
                        vprint("removing {} file due to non-existent {} file", fname, fname[:-1])
                        sftp.remove(os.path.join(dirpath, fname))
                elif not os.path.exists(os.path.join(src_dirpath, fname)):
                    sftp.remove(os.path.join(dirpath, fname))


def initial_sync(sftp, source_base_path, remote_base_path):
    vprint("syncing source to target")
    _add_source_to_target(sftp, source_base_path, remote_base_path)
    vprint("removing files/directories from target that don't exist in source")
    _remove_target_when_missing_source(sftp, source_base_path, remote_base_path)


def watch_for_changes(sftp, source_base_path, remote_base_path, batch_interval):
    vprint("watching for changes")
    queue = Queue()
    worker = SyncWorker(sftp, queue, source_base_path, remote_base_path, batch_interval)
    worker.start()

    observer = Observer()
    observer.schedule(SyncHandler(queue, source_base_path, source_ignore_patterns), source_base_path, recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("caught keyboard interrupt, stopping.")
        worker.stop()
        observer.stop()
    observer.join()
    worker.join()


def main(argv=sys.argv[1:]):
    global verbose, python_mode, dry_run, source_ignore_patterns, target_ignore_patterns, compare_by_mtime, compare_by_size
    args = docopt(__doc__, argv=argv)
    source_path = args["SOURCE"] if args["SOURCE"] else "."
    target_path = args["TARGET"]
    verbose = args["--verbose"]
    python_mode = args["--python"]
    dry_run = args["--dry-run"]
    source_ignore_patterns = args["--skip-source"]
    if not source_ignore_patterns:
        source_ignore_patterns = []
    elif not isinstance(source_ignore_patterns, (list, tuple)):
        source_ignore_patterns = [source_ignore_patterns]
    target_ignore_patterns = args["--skip-target"]
    if not target_ignore_patterns:
        target_ignore_patterns = []
    elif not isinstance(target_ignore_patterns, (list, tuple)):
        target_ignore_patterns = [target_ignore_patterns]
    sync_interval = float(args['--batch-interval'])
    compare_by_mtime = not args['--size-only']
    compare_by_size = not args['--mtime-only']

    if not os.path.exists(source_path):
        sys.stderr.write("error: source path {} does not exist, aborting.\n".format(source_path))
        sys.exit(2)

    if not os.path.isdir(source_path):
        sys.stderr.write("error: source path {} is not a directory, aborting.\n".format(source_path))
        sys.exit(2)

    sftp = SyncSFTPClient.create_sftp_connection(target_path, dry_run, not args['--no-preserve-time'],
                                                 identity_file=args['--identity'], logger_func=vprint)
    try:
        remote_path = sftp.remote_path()
    except (OSError, IOError), e:
        if e.errno == errno.ENOENT:
            sys.stderr.write("error: target path {} does not exist, aborting.\n".format(sftp.path))
            sys.exit(2)
        else:
            raise

    vprint("source path: {}", source_path)
    vprint("remote path: {}", remote_path)
    vprint("source ignore patterns: {}", source_ignore_patterns)
    vprint("target ignore patterns: {}", target_ignore_patterns)
    vprint("sync interval: {}", sync_interval)

    initial_sync(sftp, source_path, remote_path)

    if args["--watch"]:
        watch_for_changes(sftp, source_path, remote_path, sync_interval)



import os
import re
import stat
import errno
import paramiko
import getpass
from contextlib import closing


def _null_logger(*args):
    pass


class SyncSFTPClient(paramiko.SFTPClient):
    def __init__(self, user, host, port, path, dry_run, preserve_time=True, identity_file=None, 
                 logger_func=_null_logger):
        self.user = user
        self.host = host
        self.port = port
        self.path = path
        self.known_hosts = None
        self.id_files = [] if not identity_file else [identity_file]
        self.password = None
        self.dry_run = dry_run
        self.preserve_time = preserve_time
        self.logger_func = logger_func

        if not self.path:
            self.path = "~"
        self.port = int(self.port) if self.port else 22

        self._parse_ssh_config()
        if not self.user:
            default_user = getpass.getuser()
            self.user = raw_input('Username [{}]: '.format(default_user))
            if not self.user:
                self.user = default_user

        transport = paramiko.Transport((self.host, self.port))
        transport.start_client()

        self._check_known_host_keys(transport.get_remote_server_key())
        if not self._try_id_files(transport):
            if not self.password:
                self.password = getpass.getpass('Password for {}@{}: '.format(self.user, self.host))
            transport.auth_password(self.user, self.password)

        chan = transport.open_session()
        if chan is None:
            raise Exception("failed to open sftp (channel is none)")
        chan.invoke_subsystem('sftp')

        super(SyncSFTPClient, self).__init__(chan)

    def remote_path(self):
        remote_homedir = self.normalize(".")  # find the home dir
        path = self.path

        if path.startswith("~"):
            path = path.replace("~", remote_homedir, 1)
        return self.normalize(os.path.normpath(path))

    def remove(self, path):
        self.logger_func("rm remote:{}", path)
        if not self.dry_run:
            return super(SyncSFTPClient, self).remove(path)

    def mkdir(self, path):
        self.logger_func("mkdir remote:{}", path)
        if not self.dry_run:
            return super(SyncSFTPClient, self).mkdir(path)

    def put(self, src, dst):
        self.logger_func("cp {} remote:{}", src, dst)
        if not self.dry_run:
            src_stat = os.lstat(src)
            super(SyncSFTPClient, self).put(src, dst)
            if self.preserve_time:
                self.utime(dst, (src_stat.st_atime, src_stat.st_mtime))

    def mkdir_if_not_exist(self, path):
        try:
            remote_stat = self.stat(path)
            if not stat.S_ISDIR(remote_stat.st_mode):
                raise Exception("remote path {} should be a directory, but it's not.".format(path))
        except IOError, e:
            if e.errno == errno.ENOENT:
                self.mkdir(path)
            else:
                raise

    def rm(self, path):
        try:
            self.logger_func("rm remote:{}", path)
            self.remove(path)
        except IOError, e:
            if e.errno != errno.ENOENT:
                raise

    def _rm_rf_dir(self, path):
        stat_attrs = self.listdir_attr(path)
        for o in stat_attrs:
            child_path = os.path.join(path, o.filename)
            if stat.S_ISDIR(o.st_mode):
                self._rm_rf_dir(child_path)
            else:
                super(SyncSFTPClient, self).remove(child_path)

        super(SyncSFTPClient, self).rmdir(path)

    def rm_rf(self, path):
        self.logger_func("rm -rf remote:{}", path)
        if self.dry_run:
            return

        s = self.stat_or_none(path)
        if not s:
            return

        if stat.S_ISDIR(s.st_mode):
            self._rm_rf_dir(path)
        else:
            super(SyncSFTPClient, self).remove(path)

    def stat_or_none(self, path):
        try:
            return self.stat(path)
        except IOError, e:
            if e.errno != errno.ENOENT:
                raise
            return None

    def _parse_ssh_config(self):
        ssh_config_path = os.path.expanduser("~/.ssh/config")

        if not os.path.exists(ssh_config_path):
            return

        ssh_config = paramiko.SSHConfig()
        with closing(open(ssh_config_path, "r")) as f:
            ssh_config.parse(f)

        host_config = ssh_config.lookup(self.host)
        if not self.user and 'user' in host_config:
            self.user = host_config['user']
        if 'identityfile' in host_config:
            self.id_files.extend(host_config['identityfile'])
        if not self.port and 'port' in host_config:
            self.port = int(host_config['port'])
        if 'userknownhostsfile' in host_config:
            self.known_hosts = host_config['userknownhostsfile']

    def _check_known_host_keys(self, key):
        """If we have a known_hosts file, try to see that the server has the same credentials if it's there."""
        default_known_hosts = os.path.expanduser("~/.ssh/known_hosts")
        if not self.known_hosts and os.path.exists(default_known_hosts):
            self.known_hosts = default_known_hosts

        if not self.known_hosts or not os.path.exists(self.known_hosts):
            return

        known_host_key, known_host_key_type = None, None
        try:
            known_host_keys = paramiko.util.load_host_keys(os.path.expanduser(self.known_hosts))
            if self.host in known_host_keys:
                known_host_key_type = known_host_keys[self.host].keys()[0]
                known_host_key = known_host_keys[self.host][known_host_key_type]

                if known_host_key.get_name() != key.get_name() or str(known_host_key) != str(key):
                    raise Exception("Known host key verification mismatch, aborting.")
        except IOError:
            print("unable to read known hosts file {}".format(self.known_hosts))

    def _try_id_files(self, transport):
        """Try to authenticate with all the identity files at our disposal."""
        default_id = os.path.expanduser("~/.ssh/id_rsa")
        if not self.id_files and os.path.exists(default_id):
            self.id_files.append(default_id)

        for path in self.id_files:
            if not os.path.exists(path):
                continue

            try:
                pkey = paramiko.DSSKey.from_private_key_file(path)
            except paramiko.PasswordRequiredException:
                password = getpass.getpass('Password for private key {}: '.format(path))
                pkey = paramiko.DSSKey.from_private_key_file(path, password)
            except paramiko.SSHException:
                try:
                    pkey = paramiko.RSAKey.from_private_key_file(path)
                except paramiko.PasswordRequiredException:
                    password = getpass.getpass('Password for private key {}: '.format(path))
                    pkey = paramiko.RSAKey.from_private_key_file(path, password)
            except IOError:
                print("failed reading pkey file {}, skipping it.".format(path))

            try:
                transport.auth_publickey(self.user, pkey)
                return True
            except:
                pass
        return False

    @classmethod
    def _parse_target(cls, s):
        """Parse an scp-like string (e.g. user@host:port:path)"""
        match = re.match(r"(?:([^@\:\/]+)@)?([^:\/]+)(?:\:(\d+))?(?:\:(.*))?", s)
        if not match:
            raise Exception("target {} is not a valid ssh path (user@host:path|host:path|host)".format(s))
        user = match.group(1)
        host = match.group(2)
        port = match.group(3)
        path = match.group(4)
        return (user, host, port, path)

    @classmethod
    def create_sftp_connection(cls, target, dry_run=False, preserve_time=True, identity_file=None,
                               logger_func=_null_logger):
        global verbose

        user, host, port, path = cls._parse_target(target)
        return SyncSFTPClient(user, host, port, path, dry_run, preserve_time, identity_file, logger_func)
